export default `:host(.pn-loading),.pn-loading{overflow:hidden;}:host(.pn-loading):before,.pn-loading:before{position:absolute;height:100%;width:100%;content:'';z-index:1000;background-color:rgb(255, 255, 255, 0.5);border-color:lightgray;background-repeat:no-repeat;background-position:center;background-size:auto 50%;border-width:1px;cursor:progress;}:host(.pn-loading) .pn-loading-msg,.pn-loading .pn-loading-msg{color:var(--panel-on-background-color, black);font-size:2em;position:absolute;top:72%;font-size:2em;color:black;width:100%;text-align:center;}`
